part of nueli.core.config;

// Decor shortcuts
Decor get $decor => Decor();
Decor get $deco => Decor();

/// Decor ~ [BoxDecoration] configuration class.
class Decor extends Config<Decor, DecorIN, DecorOUT> {
  Decor({Key? key}) : super(key: key, fallback: DecorIN());

  /// *@varies* ~ set fill decor paint list
  Decor fill(List<DecorPaint>? hue) => resolve(hue, #paints);

  /// *@varies* ~ set single color
  Decor color(Color color) => resolve(DecorPaint(color), #paints);

  /// *@varies* ~ adds color with stop
  Decor paint(Color color, double stop) => resolve(this().paints..add(DecorPaint(color, stop)), #paints);

  /// *@varies* ~ set stroke color
  Decor stroke(double? size, [Color? color]) => resolve(size, #stroke.size).resolve(color, #stroke.color);

  /// *@varies* ~ set orientation
  Decor orient(Alignment origin, Alignment focus) => resolve(origin, #orient.origin).resolve(focus, #orient.focus);

  /// *@varies* ~ set decor style
  Decor style(DecorStyle? style) => resolve(style, #style);

  /// *@fixed* ~ set clip behaviour
  Decor clip(Clip? clip) => resolve(clip, #clip);

  /// *@shortcut* ~ set vertical orientation
  Decor get orientVertical => orient(Alignment.topCenter, Alignment.bottomCenter);

  /// *@shortcut* ~ set horizontal orientation
  Decor get orientHorizontal => orient(Alignment.centerLeft, Alignment.centerRight);

  /// *@shortcut* ~ set diagonal orientation
  Decor get orientDiagonal => orient(Alignment.topLeft, Alignment.bottomRight);

  /// *@config reference* ~ set umbra
  Decor umbra(Umbra? umbra) => resolve(umbra, #umbra);

  /// *@config reference* ~ set inkwell
  Decor inkwell(Inkwell? inkwell) => resolve(inkwell, #inkwell);

  /// *@config reference* ~ set layout
  Decor layout(Layout? layout) => resolve(layout, #layout);

  /// *@config reference* ~ set radii
  Decor radii(Radii? radii) => resolve(radii, #radii);

  /// *@from [Inkwell]* ~ set tap callback
  Decor onTap(Function()? onTap) => resolve(onTap, #onTap);

  /// *@from [Inkwell]* ~ set longPress callback
  Decor onLongPress(Function()? onLongPress) => resolve(onLongPress, #onLongPress);

  /// *@from [Layout]* ~ set size
  Decor size(double x, [double? y]) => width(x).height(y ?? x);

  /// *@from [Layout]* ~ set width
  Decor width(double width) => resolve(width, #width);

  /// *@from [Layout]* ~ set height
  Decor height(double height) => resolve(height, #height);

  /// *@from [Layout]* ~ set width scale factor
  Decor widthFactor(double factor) => resolve(factor, #widthFactor);

  /// *@from [Layout]* ~ set height scale factor
  Decor heightFactor(double factor) => resolve(factor, #heightFactor);

  /// *@from [Layout]* ~ set padding
  Decor padding(EdgeInsets? padding) => resolve(padding, #padding);

  /// *@from [Layout]* ~ set layout
  Decor margin(EdgeInsets? margin) => resolve(margin, #margin);

  /// *@from [Layout]* ~ set alignment
  Decor alignment(Alignment? alignment) => resolve(alignment, #alignment);

  /// *@from [Layout]* ~ set box constraints
  Decor constraints(BoxConstraints? constraints) => resolve(constraints, #constraints);

  /// *@from [Layout]* ~ expand size to maximum
  Decor get expand => size(double.infinity);

  /// *@from [Layout]* ~ expand width to maximum
  Decor get expandWidth => width(double.infinity);

  /// *@from [Layout]* ~ expand height to maximum
  Decor get expandHeight => height(double.infinity);

  /// *@from [Layout]* ~ align to top left corner
  Decor get alignStart => alignment(Alignment.topLeft);

  /// *@from [Layout]* ~ align to center
  Decor get alignCenter => alignment(Alignment.center);

  /// *@from [Layout]* ~ align to bottom right corner
  Decor get alignEnd => alignment(Alignment.bottomRight);
}

class DecorIN extends ConfigIN<DecorIN, DecorOUT> {
  List<DecorPaint>? paints;
  Alignment? origin, focus;
  Color? stroke;
  double? strokeSize;

  DecorStyle? style;
  Clip? clip;

  Umbra? umbra;
  Inkwell? inkwell;
  Layout? layout;
  Radii? radii;

  DecorIN() : super(#decor);

  @override
  DecorIN operator *(double operand) {
    return DecorIN().copy(this)
      ..paints = this().paints.map((e) => e * operand).toList()
      ..origin = this().origin * operand
      ..focus = this().focus * operand
      ..stroke = this().stroke * operand
      ..umbra = this().umbra * operand
      ..inkwell = this().inkwell * operand
      ..layout = this().layout * operand
      ..radii = this().radii * operand;
  }

  @override
  DecorIN operator -(DecorIN operand) {
    return DecorIN().copy(this)
      ..paints = this().paints.minus(operand().paints).toList()
      ..origin = this().origin - operand().origin
      ..focus = this().focus - operand().focus
      ..stroke = this().stroke - operand().stroke
      ..strokeSize = this().strokeSize - operand().strokeSize
      ..umbra = this().umbra - operand().umbra
      ..inkwell = this().inkwell - operand().inkwell
      ..layout = this().layout - operand().layout
      ..radii = this().radii - operand().radii;
  }

  @override
  DecorIN operator +(DecorIN operand) {
    return DecorIN().copy(this)
      ..paints = this().paints.minus(operand().paints).toList()
      ..origin = this().origin + operand().origin
      ..focus = this().focus + operand().focus
      ..stroke = this().stroke + operand().stroke
      ..strokeSize = this().strokeSize + operand().strokeSize
      ..umbra = this().umbra + operand().umbra
      ..inkwell = this().inkwell + operand().inkwell
      ..layout = this().layout + operand().layout
      ..radii = this().radii + operand().radii;
  }

  @override
  DecorIN resolve(data, tag) {
    if (data is DecorPaint && tag == #paints) paints = [data];
    if (data is List<DecorPaint> && tag == #paints) paints = data;
    if (data is Alignment && tag == #orient.origin) origin = data;
    if (data is Alignment && tag == #orient.focus) focus = data;
    if (data is Color && tag == #stroke.color) stroke = data;
    if (data is double && tag == #stroke.size) strokeSize = data;
    if (data is DecorStyle && tag == #style) style = data;
    if (data is Clip && tag == #clip) clip = data;

    if (data is Umbra && tag == #umbra) umbra = data;
    if (data is Inkwell && tag == #inkwell) inkwell = data;
    if (data is Layout && tag == #layout) layout = data;
    if (data is Radii && tag == #radii) radii = data;

    umbra = this().umbra.resolve(data, tag);
    inkwell = this().inkwell.resolve(data, tag);
    layout = this().layout.resolve(data, tag);
    radii = this().radii.resolve(data, tag);
    return super.resolve(data, tag);
  }

  @override
  DecorIN copy(iN) {
    paints = iN.paints ?? paints;
    origin = iN.origin ?? origin;
    focus = iN.focus ?? focus;
    stroke = iN.stroke ?? stroke;
    strokeSize = iN.strokeSize ?? strokeSize;
    style = iN.style ?? style;
    clip = iN.clip ?? clip;

    umbra = this().umbra.copy(iN().umbra().iN);
    inkwell = this().inkwell.copy(iN().inkwell().iN);
    radii = this().radii.copy(iN().radii().iN);
    return super.copy(iN);
  }

  @override
  DecorOUT call() => DecorOUT(this);
}

class DecorOUT extends ConfigOUT<DecorOUT, DecorIN> {
  const DecorOUT(DecorIN iN) : super(iN);

  List<DecorPaint> get paints => iN.paints ?? [];
  Alignment get origin => iN.origin ?? Alignment.centerLeft;
  Alignment get focus => iN.focus ?? Alignment.centerRight;
  Color get stroke => iN.stroke ?? palette.none;
  double get strokeSize => iN.strokeSize ?? 0;

  DecorStyle get style => iN.style ?? DecorStyle.linear;
  Clip get clip => iN.clip ?? Clip.antiAlias;

  Umbra get umbra => iN.umbra ?? $umbra;
  Inkwell get inkwell => iN.inkwell ?? $inkwell;
  Layout get layout => iN.layout ?? $layout;
  Radii get radii => iN.radii ?? $radii;

  Gradient? get toGradient {
    if (paints.isEmpty) return null;
    if (paints.isUnitary) return null;

    if (style == DecorStyle.linear) {
      return LinearGradient(
        colors: paints,
        begin: origin,
        end: focus,
        stops: paints.stops,
      );
    }
    if (style == DecorStyle.radial) {
      return RadialGradient(
        colors: paints,
        focal: origin,
        center: focus,
        stops: paints.stops,
      );
    }
    if (style == DecorStyle.sweep) {
      return SweepGradient(
        colors: paints,
        center: origin,
        stops: paints.stops,
      );
    }
    return null;
  }

  Decoration get toDecoration {
    return BoxDecoration(
      color: paints.isEmpty ? palette.none : paints.first,
      borderRadius: radii().toBorderRadius,
      gradient: toGradient,
      boxShadow: umbra().toBoxShadowDeep,
      border: (strokeSize != 0) ? Border.all(color: stroke, width: strokeSize) : null,
    );
  }

  @override
  Widget build(BuildContext context) {
    return inkwell().wrap([
      Container(
        child: child,
        constraints: layout().constraints,
        width: layout().width,
        height: layout().height,
        alignment: layout().alignment,
        padding: layout().padding,
        margin: layout().margin,
        decoration: toDecoration,
        clipBehavior: clip,
      ),
    ]);
  }
}

class DecorTween extends ConfigTween<Decor, DecorIN, DecorOUT> {
  DecorTween({required Decor begin, required Decor end}) : super(begin: begin, end: end);

  @override
  Decor get defaultConfig => Decor();
}

class DecorTransition extends ConfigTransition<Decor, DecorIN, DecorOUT> {
  const DecorTransition({Key? key, required Animation<Decor> decor, required Widget child})
      : super(key: key, config: decor, child: child);
}

class DecorPaint extends Color with Subtract<DecorPaint>, Add<DecorPaint> {
  final double stop;

  DecorPaint(Color color, [this.stop = 1]) : super(color.value);
  DecorPaint.int(int value, [this.stop = 1]) : super(value);

  DecorPaint operator *(double factor) {
    return DecorPaint.int((Color(value) * factor).value, stop * factor);
  }

  @override
  DecorPaint operator -(DecorPaint operand) {
    return DecorPaint.int(
      (Color(value) - Color(operand.value)).value,
      stop - operand.stop,
    );
  }

  @override
  DecorPaint operator +(DecorPaint operand) {
    return DecorPaint.int(
      (Color(value) + Color(operand.value)).value,
      stop + operand.stop,
    );
  }
}

enum DecorStyle { linear, radial, sweep }

extension DecorColor on Color {
  Color operator *(double factor) =>
      Color.fromRGBO((red * factor).round(), (green * factor).round(), (blue * factor).round(), opacity * factor);
  Color operator -(Color color) =>
      Color.fromRGBO(red - color.red, green - color.green, blue - color.blue, opacity - color.opacity);
  Color operator +(Color color) =>
      Color.fromRGBO(red + color.red, green + color.green, blue + color.blue, opacity + color.opacity);
}

extension on List<DecorPaint> {
  List<double> get stops {
    List<double> stops = [];
    for (DecorPaint paint in this) {
      stops.add(paint.stop);
    }
    return stops;
  }
}
