part of nueli.core.config;

// Ink shortcuts
Inkwell get $ink => Inkwell();
Inkwell get $well => Inkwell();
Inkwell get $inkwell => Inkwell();

/// ## Inkwell ~ [InkWell] & [InkResponse]
/// ## configuration class.
class Inkwell extends Config<Inkwell, InkwellIN, InkwellOUT> {
  static int get defA => palette.shade.alpha;

  Inkwell({Key? key}) : super(key: key, fallback: InkwellIN());

  /// *@varies* ~ set splash color
  Inkwell splash(Color color, [int? alpha]) => resolve(color.withAlpha(alpha ?? defA), #splash);

  /// *@varies* ~ set hover color
  Inkwell hover(Color color, [int? alpha]) => resolve(color.withAlpha(alpha ?? defA), #hover);

  /// *@varies* ~ set focus color
  Inkwell focus(Color color, [int? alpha]) => resolve(color.withAlpha(alpha ?? defA), #focus);

  /// *@varies* ~ set overlay color
  Inkwell overlay(Color color, [int? alpha]) => resolve(color.withAlpha(alpha ?? defA), #overlay);

  /// *@varies* ~ set highlight color
  Inkwell highlight(Color color, [int? alpha]) => resolve(color.withAlpha(alpha ?? defA), #highlight);

  /// *@fixed* ~ set Inkwell type
  Inkwell type(InkwellType? type) => resolve(type, #type);

  /// *@fixed* ~ set tap callback
  Inkwell onTap(Function()? onTap) => resolve(onTap, #onTap);

  /// *@fixed* ~ set long press callback
  Inkwell onLongPress(Function()? onLongPress) => resolve(onLongPress, #onLongPress);

  /// *@config reference* ~ set radii
  Inkwell radii(Radii radii) => resolve(radii, #radii);

  /// *@template* ~ create a new ripple inkwell
  factory Inkwell.ripple(Color color, [int alpha = 100]) {
    return Inkwell()
      ..splash(color, alpha)
      ..hover(color, alpha ~/ 7)
      ..focus(color, alpha ~/ 3.5)
      ..overlay(color, alpha ~/ 5)
      ..highlight(color, alpha ~/ 7);
  }

  /// *@template* ~ create a new rumble inkwell
  factory Inkwell.rumble(Color color, [int alpha = 100]) {
    return Inkwell.ripple(color, alpha)..type(InkwellType.response);
  }

  /// *@template* ~ create a new rumble inkwell
  static Inkwell get clear {
    return Inkwell()
      ..splash(palette.none, 0)
      ..hover(palette.none, 0)
      ..focus(palette.none, 0)
      ..overlay(palette.none, 0)
      ..highlight(palette.none, 0);
  }
}

class InkwellIN extends ConfigIN<InkwellIN, InkwellOUT> {
  Color? splash, hover, focus, overlay, highlight;
  Radii? radii;

  InkwellType? type;
  Function()? onTap, onLongPress;

  InkwellIN() : super(#inkwell);

  @override
  InkwellIN operator *(double operand) {
    return InkwellIN().copy(this)
      ..splash = this().splash * operand
      ..hover = this().hover * operand
      ..focus = this().focus * operand
      ..overlay = this().overlay * operand
      ..highlight = this().highlight * operand
      ..radii = this().radii * operand;
  }

  @override
  InkwellIN operator -(InkwellIN operand) {
    return InkwellIN().copy(this)
      ..splash = this().splash - operand().splash
      ..hover = this().hover - operand().hover
      ..focus = this().focus - operand().focus
      ..overlay = this().overlay - operand().overlay
      ..highlight = this().highlight - operand().highlight
      ..radii = this().radii - operand().radii;
  }

  @override
  InkwellIN operator +(InkwellIN operand) {
    return InkwellIN().copy(this)
      ..splash = this().splash + operand().splash
      ..hover = this().hover + operand().hover
      ..focus = this().focus + operand().focus
      ..overlay = this().overlay + operand().overlay
      ..highlight = this().highlight + operand().highlight
      ..radii = this().radii + operand().radii;
  }

  @override
  InkwellIN resolve(data, tag) {
    if (data is Color && tag == #splash) splash = data;
    if (data is Color && tag == #hover) hover = data;
    if (data is Color && tag == #focus) focus = data;
    if (data is Color && tag == #overlay) overlay = data;
    if (data is Color && tag == #highlight) highlight = data;

    if (data is InkwellType && tag == #type) type = data;
    if (data is Function() && tag == #onTap) onTap = data;
    if (data is Function() && tag == #onLongPress) onLongPress = data;

    if (data is Radii && tag == #radii) radii = data;

    radii = this().radii.resolve(data, tag);
    return super.resolve(data, tag);
  }

  @override
  InkwellIN copy(iN) {
    splash = iN.splash ?? splash;
    hover = iN.hover ?? hover;
    focus = iN.focus ?? focus;
    overlay = iN.overlay ?? overlay;
    highlight = iN.highlight ?? highlight;

    type = iN.type ?? type;
    onTap = iN.onTap ?? onTap;
    onLongPress = iN.onLongPress ?? onLongPress;

    radii = this().radii.copy(iN().radii().iN);
    return super.copy(iN);
  }

  @override
  InkwellOUT call() => InkwellOUT(this);
}

class InkwellOUT extends ConfigOUT<InkwellOUT, InkwellIN> {
  const InkwellOUT(InkwellIN iN) : super(iN);

  Color get splash => iN.splash ?? palette.accent.withAlpha(palette.tint.alpha);
  Color get hover => iN.hover ?? palette.accent.withAlpha(palette.tint.alpha ~/ 2);
  Color get focus => iN.focus ?? palette.accent.withAlpha(palette.tint.alpha ~/ 5);
  Color get overlay => iN.overlay ?? palette.accent.withAlpha(palette.tint.alpha ~/ 2);
  Color get highlight => iN.highlight ?? palette.accent.withAlpha(palette.tint.alpha ~/ 3);

  InkwellType get type => iN.type ?? InkwellType.well;
  Function() get onTap => iN.onTap ?? () {};
  Function() get onLongPress => iN.onLongPress ?? () {};

  Radii get radii => iN.radii ?? Radii.zero;

  bool get isNotActive {
    bool isNotActive = true;
    isNotActive = isNotActive && iN.onTap == null;
    isNotActive = isNotActive && iN.onLongPress == null;
    return isNotActive;
  }

  @override
  Widget wrap(children) {
    if (isNotActive) return super.wrap(children);
    return type == InkwellType.response
        ? InkResponse(
            child: super.wrap(children),
            highlightColor: overlay,
            overlayColor: MaterialStateProperty.all<Color>(overlay),
            splashColor: splash,
            hoverColor: hover,
            focusColor: focus,
            onTap: onTap,
            onLongPress: onLongPress,
          )
        : InkWell(
            child: super.wrap(children),
            borderRadius: radii().toBorderRadius,
            splashColor: splash,
            hoverColor: hover,
            focusColor: focus,
            overlayColor: MaterialStateProperty.all<Color>(overlay),
            highlightColor: highlight,
            onTap: onTap,
            onLongPress: onLongPress,
          );
  }

  @override
  Widget build(BuildContext context) {
    return type == InkwellType.response
        ? InkResponse(
            child: child,
            highlightColor: overlay,
            overlayColor: MaterialStateProperty.all<Color>(overlay),
            splashColor: splash,
            hoverColor: hover,
            focusColor: focus,
            onTap: onTap,
            onLongPress: onLongPress,
          )
        : InkWell(
            child: child,
            borderRadius: radii().toBorderRadius,
            splashColor: splash,
            hoverColor: hover,
            focusColor: focus,
            overlayColor: MaterialStateProperty.all<Color>(overlay),
            highlightColor: highlight,
            onTap: onTap,
            onLongPress: onLongPress,
          );
  }
}

class InkwellTween extends ConfigTween<Inkwell, InkwellIN, InkwellOUT> {
  InkwellTween({required Inkwell begin, required Inkwell end}) : super(begin: begin, end: end);

  @override
  Inkwell get defaultConfig => Inkwell();
}

class InkwellTransition extends ConfigTransition<Inkwell, InkwellIN, InkwellOUT> {
  const InkwellTransition({Key? key, required Animation<Inkwell> inkwell, required Widget child})
      : super(key: key, config: inkwell, child: child);
}

enum InkwellType { response, well }
