part of nueli.core.config;

// Radii shortcut

Radii get $radii => Radii();

Radii get $round => Radii();

/// Radii ~ [BorderRadius] configuration class.

/// All commands and fields varies and are self -

/// explanatory.

/// ```

class Radii extends Config<Radii, RadiiIN, RadiiOUT> {
  Radii({Key? key}) : super(key: key, fallback: RadiiIN());

  Radii tlX(double value) {
    return resolve(value, #tlX);
  }

  Radii tlY(double value) {
    return resolve(value, #tlY);
  }

  Radii trX(double value) {
    return resolve(value, #trX);
  }

  Radii trY(double value) {
    return resolve(value, #trY);
  }

  Radii brX(double value) {
    return resolve(value, #brX);
  }

  Radii brY(double value) {
    return resolve(value, #brY);
  }

  Radii blX(double value) {
    return resolve(value, #blX);
  }

  Radii blY(double value) {
    return resolve(value, #blY);
  }

  Radii tl(double x, [double? y]) => tlX(x).tlY(y ?? x);

  Radii tr(double x, [double? y]) => trX(x).trY(y ?? x);

  Radii br(double x, [double? y]) => brX(x).brY(y ?? x);

  Radii bl(double x, [double? y]) => blX(x).blY(y ?? x);

  Radii left(double value) => tlY(value).blY(value);

  Radii right(double value) => trY(value).brY(value);

  Radii top(double value) => tlX(value).trX(value);

  Radii bottom(double value) => blX(value).brX(value);

  factory Radii.vertical(double top, double bottom) => Radii().tl(top).tr(top).bl(bottom).br(bottom);

  factory Radii.horizontal(double left, double right) => Radii().tl(left).tr(right).bl(left).br(right);

  factory Radii.corners(double tl, double tr, double br, double bl) => Radii().tl(tl).tr(tr).br(br).bl(bl);

  factory Radii.sides(double left, double right, double top, double bottom) {
    return Radii().left(left).right(right).top(top).bottom(bottom);
  }

  factory Radii.all(double value) => Radii().left(value).right(value).top(value).bottom(value);

  static Radii get r8 => Radii.all(8);

  static Radii get r16 => Radii.all(16);

  static Radii get r24 => Radii.all(24);

  static Radii get r36 => Radii.all(36);

  static Radii get r54 => Radii.all(54);

  static Radii get r108 => Radii.all(108);

  static Radii get zero => Radii.all(0);
}

class RadiiIN extends ConfigIN<RadiiIN, RadiiOUT> {
  double? tlX;
  double? tlY;
  double? trX;
  double? trY;
  double? brX;
  double? brY;
  double? blX;
  double? blY;

  RadiiIN() : super(#radii);

  @override
  RadiiOUT call() => RadiiOUT(this);

  @override
  RadiiIN operator *(double operand) {
    tlX = this().tlX * operand;

    tlY = this().tlY * operand;

    trX = this().trX * operand;

    trY = this().trY * operand;

    brX = this().brX * operand;

    brY = this().brY * operand;

    blX = this().blX * operand;

    blY = this().blY * operand;

    return this;
  }

  @override
  RadiiIN operator -(RadiiIN operand) {
    tlX = this().tlX - operand().tlX;

    tlY = this().tlY - operand().tlY;

    trX = this().trX - operand().trX;

    trY = this().trY - operand().trY;

    brX = this().brX - operand().brX;

    brY = this().brY - operand().brY;

    blX = this().blX - operand().blX;

    blY = this().blY - operand().blY;

    return this;
  }

  @override
  RadiiIN operator +(RadiiIN operand) {
    tlX = this().tlX + operand().tlX;

    tlY = this().tlY + operand().tlY;

    trX = this().trX + operand().trX;

    trY = this().trY + operand().trY;

    brX = this().brX + operand().brX;

    brY = this().brY + operand().brY;

    blX = this().blX + operand().blX;

    blY = this().blY + operand().blY;

    return this;
  }

  @override
  RadiiIN resolve(data, Symbol tag) {
    if (data is double && tag == #tlX) tlX = data;

    if (data is double && tag == #tlY) tlY = data;

    if (data is double && tag == #trX) trX = data;

    if (data is double && tag == #trY) trY = data;

    if (data is double && tag == #brX) brX = data;

    if (data is double && tag == #brY) brY = data;

    if (data is double && tag == #blX) blX = data;

    if (data is double && tag == #blY) blY = data;

    return super.resolve(data, tag);
  }

  @override
  RadiiIN copy(iN) {
    tlX = iN.tlX ?? tlX;

    tlY = iN.tlY ?? tlY;

    trX = iN.trX ?? trX;

    trY = iN.trY ?? trY;

    brX = iN.brX ?? brX;

    brY = iN.brY ?? brY;

    blX = iN.blX ?? blX;

    blY = iN.blY ?? blY;

    return super.copy(iN);
  }
}

class RadiiOUT extends ConfigOUT<RadiiOUT, RadiiIN> {
  RadiiOUT(RadiiIN iN) : super(iN);

  double get tlX => iN.tlX ?? 0;

  double get tlY => iN.tlX ?? 0;

  double get trX => iN.trX ?? 0;

  double get trY => iN.trX ?? 0;

  double get blX => iN.blX ?? 0;

  double get blY => iN.blX ?? 0;

  double get brX => iN.brX ?? 0;

  double get brY => iN.brX ?? 0;

  BorderRadius get toBorderRadius {
    return BorderRadius.only(
      topLeft: Radius.elliptical(tlX, tlY),
      topRight: Radius.elliptical(trX, trY),
      bottomLeft: Radius.elliptical(blX, blY),
      bottomRight: Radius.elliptical(brX, brY),
    );
  }

  @override
  Widget build(context) {
    return Container(
      child: child,
      decoration: BoxDecoration(borderRadius: toBorderRadius),
      clipBehavior: Clip.antiAliasWithSaveLayer,
    );
  }
}

class RadiiTween extends ConfigTween<Radii, RadiiIN, RadiiOUT> {
  RadiiTween({required Radii begin, required Radii end}) : super(begin: begin, end: end);

  @override
  Radii get defaultConfig => Radii();
}

class RadiiTransition extends ConfigTransition<Radii, RadiiIN, RadiiOUT> {
  const RadiiTransition({Key? key, required Animation<Radii> radii, required Widget child})
      : super(key: key, config: radii, child: child);
}
