part of nueli.core.config;

// Txt shortcuts
Txt get $txt => Txt();
Txt get $text => Txt();

/// Txt ~ [Text] and [TextStyle] configuration class.
class Txt extends Config<Txt, TxtIN, TxtOUT> {
  Txt({Key? key}) : super(key: key, fallback: TxtIN());

  /// *@varies* ~ set color
  Txt color(Color color) => resolve(color, #color);

  /// *@varies* ~ set size
  Txt size(double size) => resolve(size, #size);

  /// *@fixed* ~ set text
  Txt text(String? text) => resolve(text, #text);

  /// *@fixed* ~ set font weight
  Txt weight(FontWeight? weight) => resolve(weight, #weight);

  /// *@fixed* ~ add text decoration
  Txt decoration(TextDecoration decoration) => resolve(this().decorations..add(decoration), #decorations);

  /// *@fixed* ~ set text baseline
  Txt baseline(TextBaseline? baseline) => resolve(baseline, #baseline);

  /// *@fixed* ~ set text overflow
  Txt overflow(TextOverflow? overflow) => resolve(overflow, #overflow);

  /// *@fixed* ~ set font style
  Txt style(FontStyle? style) => resolve(style, #style);

  /// *@fixed* ~ set text align
  Txt align(TextAlign? align) => resolve(align, #align);

  /// *@fixed* ~ set text span children
  Txt inline(List<Txt> children) => this.children(children);

  /// *@shortcut* ~ align to center
  Txt get alignCenter => align(TextAlign.center);

  /// *@shortcut* ~ align to center
  Txt get alignStart => align(TextAlign.start);

  /// *@shortcut* ~ align to center
  Txt get alignEnd => align(TextAlign.end);

  /// *@shortcut* ~ set font style to italics
  Txt get italics => style(FontStyle.italic);

  /// *@shortcut* ~ set font weight to bold
  Txt get bold => weight(FontWeight.bold);

  /// *@shortcut* ~ add underline to decorations
  Txt get underline => decoration(TextDecoration.underline);

  /// *@shortcut* ~ add overline to decorations
  Txt get overline => decoration(TextDecoration.overline);

  /// *@shortcut* ~ add lineThrough to decorations
  Txt get lineThrough => decoration(TextDecoration.lineThrough);

  /// *@fixed* ~ set text inkwell
  Txt inkwell(Inkwell inkwell) => resolve(inkwell, #inkwell);

  /// *@fixed* ~ set text umbra
  Txt umbra(Umbra umbra) => resolve(umbra, #umbra);
}

class TxtIN extends ConfigIN<TxtIN, TxtOUT> {
  Color? color, background;
  double? size;

  String? text;
  int? maxLines;
  FontWeight? weight;
  List<TextDecoration>? decorations;
  TextBaseline? baseline;
  TextOverflow? overflow;
  FontStyle? style;
  TextAlign? align;

  Umbra? umbra;
  Inkwell? inkwell;

  TxtIN() : super(#txt);

  @override
  TxtIN operator *(double operand) {
    return TxtIN().copy(this)
      ..color = this().color * operand
      ..background = this().backgroundColor * operand
      ..size = this().size * operand
      ..umbra = this().umbra * operand
      ..inkwell = this().inkwell * operand;
  }

  @override
  TxtIN operator -(TxtIN operand) {
    return TxtIN().copy(this)
      ..color = this().color - operand().color
      ..background = this().backgroundColor - operand().backgroundColor
      ..size = this().size - operand().size
      ..umbra = this().umbra - operand().umbra
      ..inkwell = this().inkwell - operand().inkwell;
  }

  @override
  TxtIN operator +(TxtIN operand) {
    return TxtIN().copy(this)
      ..color = this().color + operand().color
      ..background = this().backgroundColor + operand().backgroundColor
      ..size = this().size + operand().size
      ..umbra = this().umbra + operand().umbra
      ..inkwell = this().inkwell + operand().inkwell;
  }

  @override
  TxtIN resolve(data, tag) {
    if (data is Color && tag == #color) color = data;
    if (data is Color && tag == #background) background = data;
    if (data is double && tag == #size) size = data;

    if (data is String && tag == #text) text = data;
    if (data is int && tag == #maxLines) maxLines = data;
    if (data is FontWeight && tag == #weight) weight = data;
    if (data is List<TextDecoration> && tag == #decorations) decorations = data;
    if (data is TextBaseline && tag == #baseline) baseline = data;
    if (data is TextOverflow && tag == #overflow) overflow = data;
    if (data is FontStyle && tag == #style) style = data;
    if (data is TextAlign && tag == #align) align = data;

    if (data is Umbra && tag == #umbra) umbra = data;
    if (data is Inkwell && tag == #inkwell) inkwell = data;

    umbra = this().umbra.resolve(data, tag);
    inkwell = this().inkwell.resolve(data, tag);
    return super.resolve(data, tag);
  }

  @override
  TxtIN copy(TxtIN iN) {
    color = iN.color ?? color;
    background = iN.background ?? background;
    size = iN.size ?? size;

    text = iN.text ?? text;
    maxLines = iN.maxLines ?? maxLines;
    weight = iN.weight ?? weight;
    decorations = iN.decorations ?? decorations;
    baseline = iN.baseline ?? baseline;
    overflow = iN.overflow ?? overflow;
    style = iN.style ?? style;
    align = iN.align ?? align;

    umbra = this().umbra.copy(iN().umbra().iN);
    inkwell = this().inkwell.copy(iN().inkwell().iN);
    inkwell = this().inkwell.copy(iN().inkwell().iN);
    return super.copy(iN);
  }

  @override
  TxtOUT call() => TxtOUT(this);
}

class TxtOUT extends ConfigOUT<TxtOUT, TxtIN> {
  TxtOUT(TxtIN iN) : super(iN);

  Color get color => iN.color ?? palette.onBackground;
  Color get backgroundColor => iN.background ?? palette.none;
  double get size => iN.size ?? 12;

  String get text => iN.text ?? '';
  int? get maxLines => iN.maxLines;
  FontWeight get weight => iN.weight ?? FontWeight.normal;
  List<TextDecoration> get decorations => iN.decorations ?? [];

  TextBaseline? get baseline => iN.baseline;
  TextOverflow get overflow => iN.overflow ?? TextOverflow.ellipsis;
  FontStyle get style => iN.style ?? FontStyle.normal;
  TextAlign get align => iN.align ?? TextAlign.justify;

  Umbra get umbra => iN.umbra ?? $umbra;
  Inkwell get inkwell => iN.inkwell ?? $well;

  TextStyle get toTextStyle {
    return TextStyle(
      color: color,
      decoration: TextDecoration.combine(decorations),
      fontSize: size,
      shadows: umbra().toBoxShadowDeep,
      fontStyle: style,
      fontWeight: weight,
      backgroundColor: backgroundColor,
      textBaseline: TextBaseline.alphabetic,
    );
  }

  GestureRecognizer? get toRecognizer {
    if (inkwell().iN.onTap != null) {
      return TapGestureRecognizer()..onTap = () => inkwell().onTap;
    }
    if (inkwell().iN.onLongPress != null) {
      return LongPressGestureRecognizer()..onLongPress = () => inkwell().onLongPress;
    }
    return null;
  }

  @override
  Widget build(context) {
    Widget result = const SizedBox();

    final children = this.children;
    if (children.isNotEmpty && children is List<Txt>) {
      List<TextSpan> spans = [];
      for (Txt config in children) {
        spans.add(TextSpan(
          text: config.update((data) => TxtIN().copy(data))().text,
          style: config.update((data) => TxtIN().copy(iN).copy(data))().toTextStyle,
          recognizer: config.update((data) => TxtIN().copy(iN).copy(data))().toRecognizer,
        ));
      }
      result = SizedBox(
        child: Text.rich(
          TextSpan(
            text: text,
            style: toTextStyle,
            children: spans,
            recognizer: toRecognizer,
          ),
          textAlign: align,
          maxLines: maxLines,
          overflow: overflow,
          style: toTextStyle,
        ),
      );
    } else {
      result = SizedBox(
        child: Text(text, // break
            textAlign: align,
            maxLines: maxLines,
            overflow: overflow,
            style: toTextStyle),
      );
    }
    return inkwell().wrap([result, ...children]);
  }
}

class TxtTween extends ConfigTween<Txt, TxtIN, TxtOUT> {
  TxtTween({required Txt begin, required Txt end}) : super(begin: begin, end: end);

  @override
  Txt get defaultConfig => Txt();
}

class TxtTransition extends ConfigTransition<Txt, TxtIN, TxtOUT> {
  const TxtTransition({Key? key, required Animation<Txt> txt, required Widget child})
      : super(key: key, config: txt, child: child);
}
